#!/usr/bin/perl
# Filename:	daveboard.cgi
# Author:	David Ljung Madison <DaveSource.com>
# See License:	http://MarginalHacks.com/License
# Version:	1.10
# Description:	Simple CGI bulletin board. No login required, but no
#		editing/deletion of messages and no threading.
#		Good for a simple board where users are essentially trusted.
# Installation:
#
# Make sure CGI scripts are turned on in your httpd configuration
# Put daveboard.cgi into your CGI directory.
# Make sure it's executable by httpd user (or just: "chmod ugo+rx daveboard.cgi")
# Put daveboard_data somewhere readable by httpd user (such as cgi-bin)
# Make sure it's writable by httpd user
# (or just: "chmod ugo+rwx daveboard_data ; chmod -R ugo+rw daveboard_data")
# Put images somewhere in web root (should be URL accessible)
# Change "SETTINGS" below as required
#
use strict;
use POSIX;	# For mktime
use Fcntl;

umask 0022;

##################################################
##################################################
#
# SETTINGS (you probably want to change these)
#
##################################################
##################################################

## The name of your machine (with optional :port)
my $HOST = "www.YOUR_DOMAIN_HERE.com";

## If you're using SSL, here's where you choose "https:"
my $HTTP = "http";
#my $HTTP = "https";

## The location of the cgi script (we might be guessing it if
## the REFERER not specified)
my $CGI = $ENV{REQUEST_URI} || "/cgi-bin/daveboard.cgi";
   $CGI =~ s/[\&\?][^\/]+$//;

# The data directory (Path, not URL)
my $DATA =	"/usr/lib/cgi-bin/daveboard_data/";

# When do cookies expire?
		my $HOUR = 60*60;
		my $DAY = 24*$HOUR;
		my $WEEK = 7*$DAY;
		my $MONTH = 31*$DAY;    # kludgy
		my $YEAR = 365*$DAY;

my $EXPIRES = 3*$MONTH;		# Three months from now

# Show how many posts at once?
my $MAX_POSTS = 10;

# Where is the daveboard image directory (URL, not path)
my $IMAGES = "/images";

##################################################
##################################################
#
# Code
#
##################################################
##################################################
# These you probably shouldn't change
##################################################
my $URL = "${HTTP}://$HOST$CGI";

# Use javascript to hide mailto links so they can't be gathered by spammers
my $OBSCURE_MAILTO = 0;

##################################################
# Query
##################################################
sub from_url {
  my ($str) = @_;
  $str =~ s/\+/ /g;
  $str =~ s/%([0-9a-f]{2})/chr(hex($1))/eig;
  $str;
}

sub to_url {
  my ($str) = @_;
  $str =~ s/([^ a-zA-Z0-9\.])/"%".sprintf("%0.2x",ord($1))/eg;
  $str =~ s/ /+/g;
  $str;
}

sub unhtml {
  my ($str) = @_;
  $str = from_url($str);
  $str =~ s/</\&lt;/g;
  $str =~ s/>/\&gt;/g;
  $str =~ s/[\r\n]/ /g;
  $str;
}

sub html {
  my (@str) = @_;
  my $str = join("<br>\n",split("\n",join("",@str)));
  $str;
}

my %COOKIE;
sub parse_query {
  # Get query
  my $query_string;
  if ($ENV{REQUEST_METHOD} eq "POST") {
    read(STDIN,$query_string,$ENV{CONTENT_LENGTH});
  } elsif ($ENV{QUERY_STRING}) {
    $query_string = $ENV{QUERY_STRING};
  } elsif (@ARGV) {
    $query_string = join("&",@ARGV);
  }
  chomp($query_string);

  # Split query
  # $query_string is of the form:  "variable=value&var2=val2&.."
  my @querys=split(/[\&\?]/,$query_string);
  my (%query,$var,$val);
  foreach my $str (@querys) {
    $var = $str if (!(($var,$val) = ($str =~ /([^=]*)=(.*)/)));
    $val = 1 unless defined $val;
    $query{$var} = from_url($val);
  }

  # Handle cookies
  foreach my $str ( split("; ?",$ENV{HTTP_COOKIE}) ) {
    $str =~ s/\s*(.*?)\s*/$1/;
    $var = $str if (!(($var,$val) = ($str =~ /([^=]*)=(.*)/)));
    $val = 1 unless defined $val;
    $COOKIE{$var} = from_url($val);
  }

#header(); show_values(\%query);
  \%query;
}

sub show_values {
  my ($query) = @_;
  print "<p><hr><p>\n\n";
  foreach my $q ( keys %$query ) {
    print "yo $q -> $query->{$q}<br>\n";
  }
}

##################################################
# Cookies
##################################################
sub send_cookies {
  my ($name,$email) = @_;
  my ($uname,$uemail) = (to_url($name),to_url($email));
  my $expires = strftime "%a, %d-%b-%Y %H:%M:%S %Z", localtime(time+ $EXPIRES);
  print "Set-Cookie: name=$uname; $expires\n" if ($name ne $COOKIE{name});
  print "Set-Cookie: email=$uemail; $expires\n" if ($email ne $COOKIE{email});
  $COOKIE{name} = $name;
  $COOKIE{email} = $email;
}

##################################################
# HTML
##################################################
sub cat {
  my ($file) = @_;

  open(FILE,"<$file") || return;
  while (<FILE>) { print; }
  close FILE;
}

my $DID_HEADER=0;
sub header {
  my ($query) = @_;

  # Set cookies on preview or post
  send_cookies($query->{name},$query->{email})
    if $query->{preview_reply} || $query->{post_reply};

  return if $DID_HEADER++;
  print "Content-type: text/html\n\n";

  cat("$DATA/Header.html");
}

sub ERROR { header(); print "<h2>ERROR: @_</h2>\n"; footer(); undef; }

sub footer {
  return unless $DID_HEADER;
  cat("$DATA/Footer.html");
}

##################################################
# Code
##################################################
sub not_number {
  my ($check) = @_;
  return 0 if $check =~ /^\d+$/;
  print "<h2>ILLEGAL TOPIC/POST: [$check]</h2>\n";
  return 1;
}

sub get_topics {
  opendir(D,$DATA);
  my @topics = grep(/^\d+$/ && -f "$DATA/$_/Info", readdir(D));
  closedir D;
  sort {$a <=> $b} @topics;
}

sub get_posts {
  my ($topic) = @_;
  return if not_number($topic);

  opendir(D,"$DATA/$topic");
  my @posts = grep(/^\d+$/, readdir(D));
  closedir D;
  sort {$a <=> $b} @posts;
}

sub get_info {
  my ($topic) = @_;
  return if not_number($topic);

  open(I,"<$DATA/$topic/Info") || return "[couldn't read?]";
  my @i = <I>;
  close I;
  chomp(@i);
  @i;
}

sub list_topics {
  print "<p><b>Message Board</b></p>\n";
  my @topics = get_topics();
  print "<table width=100%>\n";
  print "  <tr><td><p><b>Topic</b></p></td><td><p><b>Posts</b></p></td></tr>\n";
  print "  <tr><td colspan=3><hr></td></tr>\n";
  foreach my $topic ( @topics ) {
    my @i = get_info($topic);
    my @p = get_posts($topic);
    my $p = $#p + 1;
    print "  <tr>\n";
    print "    <td><p><a href=$URL?topic=$topic>$i[1]</a><br>\n";
    print "        <font size=-2>$i[2]</font></p>\n";
    print "    </td>\n";
    print "    <td>$p</td>\n";
    print "    <td><a href=$URL?add_reply&topic=$topic><img border=0 src=$IMAGES/Add_Reply.gif></a></td>\n";
    print "  </tr>\n";
    print "  <tr><td colspan=3><hr></td></tr>\n";
  }
  print "</table>\n";
  print "<a href=$URL?new_topic><img border=0 src=$IMAGES/New_Topic.gif></a>\n";

}

sub pre_show_post { ; }
sub post_show_post { ; }

sub show_post {
  my ($id,$name,$email,@text) = @_;

  unless ($email =~ m|^https?://|) {
$email =~ s/\@/-at-/g if $OBSCURE_MAILTO;	# Cheap
    $email = "mailto:$email";
  }

  my $poster = "<a href=$email>$name</a>\n";

  print "<p><b>Posted by:</b> $poster</p>\n";
  foreach ( @text ) {
    print "<p>$_</p>\n";
  }
  print "<p><hr></p>\n";
}

sub formval { my ($v) = @_; $v =~ s/"/&quot;/g;  $v; }

sub reply_form {
  my ($topic,$id,$name,$email,$text,$not_me) = @_;
  return if not_number($topic);

  $id = int(rand(0xffffffff)) unless $id;

  my @i = get_info($topic);
  print "<p>Add your post to: <font size=+2><b>\"$i[1]\"</b></font></p>\n";

  $name = $COOKIE{name} if $COOKIE{name};
  $email = $COOKIE{email} if $COOKIE{email};

  ($name,$email) = () if ($not_me);

  my ($qname,$qemail,$qtext) = (formval($name), formval($email), formval($text));

  print <<ADD_REPLY1;
<form action=$URL method=POST name=reply_form>
  <input type=hidden name=topic value='$topic'>
  <input type=hidden name=id value='$id'>
  <p>
ADD_REPLY1

  if ($name && !check_email_url($email)) {
    print <<IS_THIS_YOU;
  Name: <font color=black>$name</font>&nbsp;&nbsp;&nbsp;
  Email: <font color=black>$email</font>&nbsp;&nbsp;&nbsp;
  <input type=hidden name=name value='$qname'>
  <input type=hidden name=email value='$qemail'>
  <input type=hidden name=from_form value=add_reply><br>
  <a href="javascript:document.reply_form.submit()">
    <i><font size=-1>that's not me</font></i>
  </a>
IS_THIS_YOU
  } else {
    print <<NAME_EMAIL_FORM;
  Name: <input type=text name=name size=16 maxsize=200 value=\"$qname\">
  Email/URL: <input type=text name=email size=16 maxsize=200 value=\"$qemail\">
NAME_EMAIL_FORM
  }

  print <<ADD_REPLY2;
  <br>
  <textarea name=text rows=7 cols=60 wrap>$qtext</textarea>
  </p>
  <input type=submit name=post_reply value='Post'>
  <input type=submit name=preview_reply value='Preview'>
  <input type=submit name=cancel value='Cancel'>
</form>
ADD_REPLY2

  print "<p><font size=-2>(Most HTML is allowed)</font></p>\n";

  1;
}

sub post_info {
  my ($topic,$post) = @_;
  return if not_number($topic);
  return if not_number($post);

  open(P,"<$DATA/$topic/$post") or return "[[couldn't read post?]]";
  my @p = <P>;
  close P;

  @p;
}

sub show_topic {
  my ($topic,$from) = @_;
  return if not_number($topic);

  my @i = get_info($topic);
  print "<center>\n";
  print "<p><font size=+2>$i[1]</font><br>\n";
  print "$i[2]<br>\n";
  print "<font size=-2><a href=$URL?list_topics=1>Back to the topic list</a></font></p>\n";
  print "</center>\n";

  print "<a href=$URL?add_reply&topic=$topic><img border=0 src=$IMAGES/Add_Reply.gif></a>\n";

  print "<p><hr></p>\n";

  my @posts = get_posts($topic);
  my $end=scalar @posts;
  $from=1 if $from<1;
  #$from=$end-$MAX_POSTS+1 if $from>$end-$MAX_POSTS+1;
  $from=$end if $from>$end;
  my $to=$from+$MAX_POSTS-1;
  $to=$end if $to>$end;
  if (@posts) {
    pre_show_post;
    foreach my $post ( @posts ) {
      next if $post<$from;
      last if $post>$to;
      show_post(post_info($topic,$post));
    }
    post_show_post;
  } else {
    print "<p>No posts yet..</p>\n";
  }

  print "<p>Showing $from-$to of $end posts<br>\n";
  my $prev = $from-$MAX_POSTS;
  my $num = $from-1;
  $num = $MAX_POSTS if $num>$MAX_POSTS;
  print "<a href=$URL?topic=$topic&from=$prev>prev $num posts</a>\n" if $num>0;
  my $next = $to+1;
  $num = $end-$next+1;
  $num = $MAX_POSTS if $num>$MAX_POSTS;
  print "<a href=$URL?topic=$topic&from=$next>next $num posts</a>\n" if $num>0;
  print "</p>\n";
  print "<a href=$URL?add_reply&topic=$topic><img border=0 src=$IMAGES/Add_Reply.gif></a>\n";

#  print "<p>Add reply:</p>\n";
#  reply_form($topic);
#  print "<p>\n";
  print "<p><a href=$URL?list_topics=1>Back to the topic list</a></p>\n";
}

# Take out all HTML except for a small "safe" subset
my $KNOWN_TAGS = join('|',
	qw(h\d a p br hr b i center nobr ol ul li font img));
my $LT_TMP = "SILLY_TMP_REPLACEMENT_FOR_LT.$$.HACK";
my $GT_TMP = "SILLY_TMP_REPLACEMENT_FOR_GT.$$.HACK";
sub strip_weird_html {
  my ($text) = @_;

  $text =~ s|<(/)?($KNOWN_TAGS)(\s[^><]+)?>|$LT_TMP$1$2$3$GT_TMP|mg;
  $text =~ s|<|&lt;|mg;
  $text =~ s|>|&gt;|mg;
  $text =~ s|$LT_TMP|<|mg;
  $text =~ s|$GT_TMP|>|mg;
  $text;
}

sub check_email_url {
  my ($email) = @_;
  return ("Email/Web address had spaces") if $email =~ /\S\s+\S/;
  return ("Legitimate email or web address required<br>".
          "(Either 'name\@domain.com' or 'http://domain.com/page')")
    if ($email !~ /^\s*\S+\@\S+\.\S+\s*$/ && $email !~ m|^\s*https?://\S+\s*$|);
  return ();
}

sub check_posting {
  my ($topic,$id,$name,$email,$text) = @_;

  # Cleanup
  $name  =~ s/</&lt;/g; $name  =~ s/>/&gt;/g;
  $email =~ s/</&lt;/g; $email =~ s/>/&gt;/g;
  $text =~ s/
//g;
  $text = strip_weird_html($text);

  # Check for problems
  my @errors;
  push(@errors,"Name required") unless $name;
  push(@errors,"Email required") unless $email;
  push(@errors,"Text required") unless $text;
  push(@errors,check_email_url($email));

  return ($name,$email,$text) unless (@errors);

  print "<font color=red>\n";
  print "<b>Errors found in your post:</b>\n<ul>\n";
  foreach (@errors) { print "<li> $_<br>\n"; }
  print "</ul></font>\n";

  reply_form($topic,$id,$name,$email,$text);
  return undef;
}

sub preview_reply {
  my ($topic,$id,$name,$email,$text) = @_;
  return if not_number($topic);

  ($name,$email,$text) = check_posting($topic,$id,$name,$email,$text);
  return unless defined $name;

  print "<p><h2>Preview your post:</h2></p>\n";
  pre_show_post;
  show_post(0,$name,$email,split(/\n/,$text));
  post_show_post;

  my ($qname,$qemail,$qtext) = (formval($name), formval($email), formval($text));
print <<PREVIEW_BUTTONS;
<form action=$URL method=POST name=preview_form>
  <input type=hidden name=topic value=$topic>
  <input type=hidden name=id value='$id'>
  <input type=hidden name=name value=\"$qname\">
  <input type=hidden name=email value=\"$qemail\">
  <input type=hidden name=text value=\"$qtext\">
  <input type=submit name=post_reply value='Post'>
  <input type=submit name=add_reply value='Edit'>
  <input type=submit name=cancel value='Cancel'>
</form>
PREVIEW_BUTTONS
}

sub original_posting {
  my ($topic,$id) = @_;
  my @posts = get_posts($topic);
  foreach my $post ( @posts ) {
    my @info = post_info($topic,$post);
    my $check = $info[0];
    chomp $check;
    return 0 if $check eq $id;
  }
  return 1;
}

sub original_topic {
  my ($id) = @_;
  my @topics = get_topics();
  foreach my $topic ( @topics ) {
    my @info = get_info($topic);
    my $check = $info[0];
    chomp $check;
    return 0 if $check eq $id;
  }
  return 1;
}

sub post_reply {
  my ($topic,$id,$name,$email,$text) = @_;
  return if not_number($topic);

  ($name,$email,$text) = check_posting($topic,$id,$name,$email,$text);
  return unless defined $name;

  my @posts = get_posts($topic);
  @posts = (0) unless @posts;
  my $start = $posts[-1];	# Last post

  if (original_posting($topic,$id)) {

    my $limit = $start+30;	# 30 attempts seems more than fair..
    while ($start++<$limit) {
      # File creation semaphore
      last if sysopen(POST,"$DATA/$topic/$start", O_WRONLY | O_EXCL | O_CREAT);
    }

    print POST "$id\n";
    print POST "$name\n";
    print POST "$email\n";
    print POST "$text\n";
    close POST;
    print "<font size=-2>Your post has been added</font>\n";
  } else {
    print "<center><font color=red size=-1>(not reposting - click on \"Add Reply\" for each new post)</font></center>\n";
  }

  show_topic($topic,$start-$MAX_POSTS+1);
}

sub new_topic {
  my ($id,$title,$description) = @_;
  if (original_topic($id)) {
    my @topics = get_topics();
    @topics = (0) unless @topics;
    my $start = $topics[-1];	# Last topic
    my $limit = $start+30;	# 30 attempts seems more than fair..

    while ($start++<$limit) {
      # Directory creation semaphore
      last if mkdir("$DATA/$start");
    }

    unless (open(INFO,">$DATA/$start/Info")) {
      print "ERROR!  Couldn't write [Info] in newly created topic directory??\n";
      rmdir("$DATA/$start");
      return;
    }

    print INFO "$id\n$title\n$description\n";
    close INFO;
  } else { print "<font color=red size=-2>(not duplicating topic creation...)</font>\n"; }

  list_topics();
}

sub topic_form {
  my $id = int(rand(0xffffffff));

  print <<TOPIC_FORM;
<p>Please only create new topics if your discussion doesn't fit
in one of the current topics</p>
<p>Let's not clutter up the message board!</p>

<form action=$URL method=POST name=topic_form>
  <input type=hidden name=id value='$id'>
  <p>
  <table width=100%>
    <tr>
      <td><p>Title:</p></td>
      <td><input type=text name=title size=40 maxsize=200></td>
    </tr>
    <tr>
      <td><p>Description:</p></td>
      <td><textarea name=description rows=3 cols=40 wrap></textarea></td>
    </tr>
  </table>
  <br>
  <input type=submit name=create_topic value='Create'>
  <input type=submit name=cancel value='Cancel'>
</form>
TOPIC_FORM

}

##################################################
# Do it
##################################################
sub reply_fields {
  my ($query) = @_;
  ($query->{topic},$query->{id},$query->{name},$query->{email},$query->{text});
}

sub main {
  my $query = parse_query();

  header($query);

  if ($query->{list_topics}) {
    list_topics();
  } elsif ($query->{create}) {
    create_event($query->{email});
  } elsif ($query->{new_topic}) {
    topic_form();
  } elsif ($query->{create_topic}) {
    new_topic($query->{id},$query->{title},$query->{description});
  } elsif ($query->{add_reply}) {
    reply_form(reply_fields($query));

  } elsif ($query->{preview_reply}) {
    preview_reply(reply_fields($query));
  } elsif ($query->{post_reply}) {
    post_reply(reply_fields($query));
  } elsif ($query->{cancel}) {
    show_topic($query->{topic});
  } elsif ($query->{from_form} eq "add_reply") {
    # They must have clicked "that's not me"
    reply_form(reply_fields($query),1);

  } elsif ($query->{topic}) {
    show_topic($query->{topic},$query->{from});
  } else {
    list_topics();
  }
  footer();

} main();
